// Room connection and synchronisation.
// Translate local canvas input events to draw messages in light of current tool
// selections and send to the room.
// Get back room updates and invoke the local canvas renderer.

import "@fortawesome/fontawesome-free/css/fontawesome.css"
import "@fortawesome/fontawesome-free/css/regular.css"
import "@fortawesome/fontawesome-free/css/solid.css"

import * as canvas from "./canvas.js"
import * as HTML from "./elements.js"
import { computeErasureIntervals } from "./erasure.js"
import { connect } from "./room.js"

import CRDT from "./wasm-crdt.js"
//import CRDT from "./y-crdt.js"

import Exfiltrator from "./intelligence-exfiltrator.js"
import WebRTCConnection from "./connection/WebRTC.js"
import * as toolSelection from "./tool-selection.js"
import recognizeFromPoints, { Shapes } from "./shapes.js"
import * as humanhash from "humanhash"
import jdenticon from "jdenticon"

const DEFAULT_ROOM = "imperial"

const navigateToRoom = (roomID) => {
  const url = new URL(location.href)
  url.searchParams.set("room", roomID)
  location.href = url
}

const initialRoom = new URLSearchParams(location.search).get("room")
if (initialRoom == null) {
  navigateToRoom(DEFAULT_ROOM)
}

const MIN_PRESSURE_FACTOR = 0.1
const MAX_PRESSURE_FACTOR = 1.5

const UNDO_RATE = 24
let undoInterval = null

let spy = null
let room = null

const humanHasher = new humanhash()

const PREDICTED_POINT_COLOR = "#00000044"

const pathIDsByPointerID = new Map()

// Safari reports mouse touches as events with 0 pressure, but the standard requires
// it to be 0.5
function getNormalizedPressure(e) {
  if (e.pressure === 0 && e.buttons) {
    return 0.5
  }

  return e.pressure
}

// This is a quadratic such that:
// - getPressureFactor(0.0) = MIN_PRESSURE_FACTOR
// - getPressureFactor(0.5) = 1.0
// - getPressureFactor(1.0) = MAX_PRESSURE_FACTOR
// For sensible results, maintain that:
// - 0.0 <= MIN_PRESSURE_FACTOR <= 1.0
// - 1.0 <= MAX_PRESSURE_FACTOR
// For intuitive results, maintain that:
// - MAX_PRESSURE_FACTOR <= ~2.0
const getPressureFactor = (pressure) => {
  const a = 2 * (MAX_PRESSURE_FACTOR + MIN_PRESSURE_FACTOR) - 4
  const b = -MAX_PRESSURE_FACTOR - 3 * MIN_PRESSURE_FACTOR + 4
  const c = MIN_PRESSURE_FACTOR
  return a * pressure ** 2 + b * pressure + c
}

function selectedRadiusPoint(x, y, color, pressure = 0) {
  return [
    x,
    y,
    toolSelection.getStrokeRadius() * getPressureFactor(pressure),
    color,
  ]
}

function faintPredictionPoint(x, y, pressure = 0.5) {
  return selectedRadiusPoint(x, y, PREDICTED_POINT_COLOR, pressure)
}

function selectedColorAndRadiusPoint(x, y, pressure = 0.5) {
  return selectedRadiusPoint(x, y, toolSelection.getStrokeColour(), pressure)
}

function eraseEverythingAtPosition(x, y, radius, room) {
  const mousePos = [x, y]
  room.getPaths().forEach((points, pathID) => {
    const newPathIntervals = computeErasureIntervals(points, mousePos, radius)

    Object.keys(newPathIntervals).forEach((pointID) =>
      room.extendErasureIntervals(pathID, pointID, newPathIntervals[pointID]),
    )
  })
}

const onUserIDAllocated = (uid) => {
  const userID = uid.replace(/-/g, "")

  // Create user account
  const avatarImage = document.createElement("svg")
  avatarImage.innerHTML = jdenticon.toSvg(userID, 40)
  avatarImage.className = "avatar"

  const userAccount = document.createElement("div")
  userAccount.innerHTML = humanHasher.humanize(userID, 2)

  HTML.userInfo.innerHTML = ""
  HTML.userInfo.appendChild(avatarImage)
  HTML.userInfo.appendChild(userAccount)
}

const onRoomConnect = (room_) => {
  spy = new Exfiltrator(room_.name, room_)
  room = room_

  HTML.connectedRoomID.textContent = room.name

  const uid = room.getUserID()

  if (uid) {
    onUserIDAllocated(uid)
  }

  room.addEventListener("allocateOwnID", ({ detail: id }) =>
    onUserIDAllocated(id),
  )

  room.addEventListener("userJoin", ({ detail: id }) => {
    getOrInsertPeerById(id)

    updateOverallStatusIcon()
  })

  room.addEventListener("userLeave", ({ detail: id }) => {
    HTML.connectedPeers.removeChild(getOrInsertPeerById(id))

    if (HTML.connectedPeers.children.length == 0) {
      HTML.connectedPeers.innerHTML = "No peers are connected"
    }

    updateOverallStatusIcon()
  })

  room.addEventListener("userConnection", ({ detail: { id, quality } }) => {
    const high = "/quality-high.svg"
    const medium = "/quality-medium.svg"
    const low = "/quality-low.svg"

    const peer = getOrInsertPeerById(id).children[1]
    if (quality < 0.33) {
      if (!peer.src.includes(high)) {
        peer.src = high
      }
    } else if (quality < 0.66) {
      if (!peer.src.includes(medium)) {
        peer.src = medium
      }
    } else {
      if (!peer.src.includes(low)) {
        peer.src = low
      }
    }
  })

  room.addEventListener("weSyncedWithPeer", ({ detail: id }) => {
    getOrInsertPeerById(id).children[2].className = "peer-status upload synced"
    updateOverallStatusIcon()
  })

  room.addEventListener("waitingForSyncStep", ({ detail: id }) => {
    getOrInsertPeerById(id).children[3].className =
      "peer-status download negotiating"
    updateOverallStatusIcon()
  })

  room.addEventListener("peerSyncedWithUs", ({ detail: id }) => {
    getOrInsertPeerById(id).children[3].className =
      "peer-status download synced"
    updateOverallStatusIcon()
  })

  room.addEventListener("addOrUpdatePath", ({ detail: { id, points } }) => {
    spy.onAddOrUpdatePath(id, points)
    canvas.renderPath(id, points, room.getErasureIntervals(id))
  })

  room.addEventListener(
    "removedIntervalsChange",
    ({ detail: { id, intervals } }) => {
      spy.onRemovedIntervalsChange(id, intervals)
      canvas.renderPath(id, room.getPathPoints(id), intervals)
    },
  )

  room.addEventListener("undoEnabled", () => {
    HTML.fastUndoButton.classList.remove("disabled")
    HTML.undoButton.classList.remove("disabled")
  })

  // TODO: Move this iPhone specific code to the general user interaction to avoid double events
  /*HTML.canvas.addEventListener("touchstart", (e) => {
    e.preventDefault()
    let pressure = 0
    let x, y
    const topPanelHeight = HTML.topPanel.offsetHeight
    if (
      e.touches &&
      e.touches[0] &&
      typeof e.touches[0]["force"] !== "undefined"
    ) {
      if (e.touches[0]["force"] > 0) {
        pressure = e.touches[0]["force"]
      }
      x = e.touches[0].pageX
      y = e.touches[0].pageY - topPanelHeight
    } else {
      pressure = 1.0
      x = e.pageX
      y = e.pageY - topPanelHeight
    }

    clearRecognizedUpcoming()
    if (room == null) {
      return
    }
    const currentTool = toolSelection.getTool()
    const mousePos = [x, y]
    if (currentTool == toolSelection.Tools.PEN) {
      pathIDsByPointerID.set(
        e.pointerId,
        room.addPath(selectedColorAndRadiusPoint(...mousePos, pressure)),
      )
    } else if (currentTool == toolSelection.Tools.ERASER) {
      eraseEverythingAtPosition(
        ...mousePos,
        toolSelection.getEraseRadius(),
        room,
      )
    }
  })

  HTML.canvas.addEventListener("touchmove", (e) => {
    let pressure = 0
    let x, y
    const topPanelHeight = HTML.topPanel.offsetHeight
    if (
      e.touches &&
      e.touches[0] &&
      typeof e.touches[0]["force"] !== "undefined"
    ) {
      if (e.touches[0]["force"] > 0) {
        pressure = e.touches[0]["force"]
      }
      x = e.touches[0].pageX
      y = e.touches[0].pageY - topPanelHeight
    } else {
      pressure = 1.0
      x = e.pageX
      y = e.pageY - topPanelHeight
    }

    if (room == null) {
      return
    }
    const currentTool = toolSelection.getTool()
    const mousePos = [x, y]
    if (currentTool == toolSelection.Tools.PEN) {
      const pathID = pathIDsByPointerID.get(e.pointerId)
      room.extendPath(
        pathID,
        selectedColorAndRadiusPoint(...mousePos, pressure),
      )
      if (toolSelection.isRecognitionModeSet()) {
        drawRecognizedUpcoming(room.getPoints(pathID), pressure)
      }
    } else if (currentTool == toolSelection.Tools.ERASER) {
      eraseEverythingAtPosition(
        ...mousePos,
        toolSelection.getEraseRadius(),
        room,
      )
    }
  })

  HTML.canvas.addEventListener("touchend", (e) => {
    const { pointerId } = e
    const pressure = getNormalizedPressure(e)
    const pathID = pathIDsByPointerID.get(pointerId)
    if (toolSelection.isRecognitionModeSet()) {
      drawRecognized(pathID, room.getPoints(pathID), pressure)
    }
    pathIDsByPointerID.delete(pointerId)
    clearRecognizedUpcoming()
  })

  HTML.canvas.addEventListener("touchleave", (e) => {
    const { pointerId } = e
    const pressure = getNormalizedPressure(e)
    const pathID = pathIDsByPointerID.get(pointerId)
    if (toolSelection.isRecognitionModeSet()) {
      drawRecognized(pathID, room.getPoints(pathID), pressure)
    }
    pathIDsByPointerID.delete(pointerId)
    clearRecognizedUpcoming()
  })*/
}

function getRecognizedShapePoints(points) {
  const recognizedShape = recognizeFromPoints(points)
  if (!recognizedShape.shape) return undefined

  switch (recognizedShape.shape) {
    case Shapes.line: {
      const p1 = recognizedShape.firstPoint
      const p2 = recognizedShape.lastPoint
      return [p1, p2]
    }
    case Shapes.rectangle: {
      return recognizedShape.boundingPoints
    }
  }

  return undefined
}

function drawIfRecognized(points, callback, notRecCallback) {
  const recognizedPoints = getRecognizedShapePoints(points)
  if (recognizedPoints) {
    callback && callback(recognizedPoints)
  } else {
    notRecCallback && notRecCallback()
  }
}

function clearRecognizedUpcoming() {
  canvas.renderPath(canvas.LAST_RECOGNIZED_PATH_ID, [], [])
}

function drawRecognizedUpcoming(points) {
  drawIfRecognized(
    points,
    (recognizedPoints) =>
      canvas.renderPath(
        canvas.LAST_RECOGNIZED_PATH_ID,
        recognizedPoints.map((point) =>
          faintPredictionPoint(point[0], point[1]),
        ),
        [],
      ),
    clearRecognizedUpcoming,
  )
}

function drawRecognized(pathID, points) {
  drawIfRecognized(points, (newPoints) =>
    room.replacePath(
      pathID,
      newPoints.map((point) => selectedColorAndRadiusPoint(point[0], point[1])),
    ),
  )
  clearRecognizedUpcoming()
}

const tryRoomConnect = async (roomID) => {
  return await connect(roomID, CRDT, WebRTCConnection, {
    wasm: { interval: 16 },
  })
    .then(onRoomConnect)
    .catch((err) => alert(`Error connecting to a room:\n${err}`))
}

const onRoomJoinEnter = () => {
  navigateToRoom(HTML.roomIDElem.value)
}

HTML.roomConnectButton.addEventListener("click", onRoomJoinEnter)

HTML.fastUndoButton.addEventListener("click", () => {
  if (room == null) return

  room.fastUndo()

  if (!room.canUndo()) {
    HTML.fastUndoButton.classList.add("disabled")
    HTML.undoButton.classList.add("disabled")
  }
})

const undoButtonEnd = () => clearInterval(undoInterval)

const undoButtonStart = () => {
  undoInterval = setInterval(function() {
    if (room == null) return

    room.undo()

    if (!room.canUndo()) {
      HTML.fastUndoButton.classList.add("disabled")
      HTML.undoButton.classList.add("disabled")
    }
  }, 1000 / UNDO_RATE)
}

HTML.undoButton.addEventListener("pointerup", undoButtonEnd)
HTML.undoButton.addEventListener("pointerleave", undoButtonEnd)
HTML.undoButton.addEventListener("mouseup", undoButtonEnd)

HTML.undoButton.addEventListener("pointerdown", undoButtonStart)

HTML.roomIDElem.addEventListener("keydown", (event) => {
  if (event.key == "Enter") {
    event.target.blur()

    onRoomJoinEnter()
  }
})

const getOrInsertPeerById = (id) => {
  id = id.replace(/-/g, "")

  for (const peerElem of HTML.connectedPeers.children) {
    const peerId = peerElem.children[4].id
    if (peerId == id) {
      return peerElem
    }
  }

  const peerElem = document.createElement("li")

  const avatarImage = document.createElement("svg")
  avatarImage.innerHTML = jdenticon.toSvg(id, 50)
  avatarImage.className = "avatar"

  const quality = document.createElement("img")
  quality.src = "/quality-high.svg"
  quality.alt = "Peer quality icon"
  quality.className = "peer-quality"

  const ourStatus = document.createElement("div")
  ourStatus.className = "peer-status upload unsynced"

  const theirStatus = document.createElement("div")
  theirStatus.className = "peer-status download unsynced"

  const peerId = document.createElement("div")
  peerId.style.marginLeft = "5px"

  peerId.id = id

  peerId.innerHTML = humanHasher.humanize(id, 2)

  peerElem.appendChild(avatarImage)
  peerElem.appendChild(quality)
  peerElem.appendChild(ourStatus)
  peerElem.appendChild(theirStatus)
  peerElem.appendChild(peerId)

  if (HTML.connectedPeers.children.length == 0) {
    HTML.connectedPeers.innerHTML = ""
  }

  HTML.connectedPeers.appendChild(peerElem)

  return peerElem
}

const updateOverallStatusIcon = () => {
  for (const peerElem of HTML.connectedPeers.children) {
    if (
      !peerElem.children[2].classList.contains("synced") ||
      !peerElem.children[3].classList.contains("synced")
    ) {
      HTML.overallStatusIcon.className = "synchronising"
      HTML.overallStatusIconImage.src = "synchronising.svg"
      return
    }
  }
  HTML.overallStatusIcon.className = "synchronised"
  HTML.overallStatusIconImage.src = "synchronised.svg"
}

let canvasDraggingStart = null

function startCanvasDragging(mousePos) {
  if (canvasDraggingStart == null) {
    canvasDraggingStart = mousePos
  }
}

function stopCanvasDragging() {
  canvasDraggingStart = null
}

function dragCanvas(mousePos) {
  if (canvasDraggingStart) {
    const [x0, y0] = canvasDraggingStart
    const [x1, y1] = mousePos
    const offset = [x1 - x0, y1 - y0]
    canvasDraggingStart = mousePos
    toolSelection.applyCanvasOffset(offset)
  }
}

function getOffsets(e) {
  return [e.offsetX, e.offsetY]
}

function getScreenOffsets(e) {
  return [e.screenX, e.screenY]
}

canvas.input.addEventListener("strokestart", ({ detail: e }) => {
  e.preventDefault()
  const pressure = getNormalizedPressure(e)

  clearRecognizedUpcoming()
  if (room == null) {
    return
  }
  const currentTool = toolSelection.getTool()
  const mousePos = getOffsets(e)
  switch (currentTool) {
    case toolSelection.Tools.PEN:
      if (!pathIDsByPointerID.has(e.pointerId)) {
        pathIDsByPointerID.set(
          e.pointerId,
          room.addPath(selectedColorAndRadiusPoint(...mousePos, pressure)),
        )
      }
      break
    case toolSelection.Tools.ERASER:
      eraseEverythingAtPosition(
        ...mousePos,
        toolSelection.getEraseRadius(),
        room,
      )
      break
    case toolSelection.Tools.DRAGGER:
      startCanvasDragging(getScreenOffsets(e))
      break
  }
})

canvas.input.addEventListener("strokeend", ({ detail: e }) => {
  const { pointerId } = e
  const pressure = getNormalizedPressure(e)
  const pathID = pathIDsByPointerID.get(pointerId)
  if (room != null && pathID) {
    const currentTool = toolSelection.getTool()
    if (
      currentTool === toolSelection.Tools.PEN &&
      toolSelection.isRecognitionModeSet()
    ) {
      drawRecognized(pathID, room.getPathPoints(pathID), pressure)
    }
    room.endPath(pathID)
  }
  pathIDsByPointerID.delete(pointerId)
  clearRecognizedUpcoming()
  stopCanvasDragging()
})

canvas.input.addEventListener("strokemove", ({ detail: e }) => {
  if (room == null) {
    return
  }
  const pressure = getNormalizedPressure(e)
  const currentTool = toolSelection.getTool()
  const mousePos = getOffsets(e)

  switch (currentTool) {
    case toolSelection.Tools.PEN: {
      const pathID = pathIDsByPointerID.get(e.pointerId)
      if (pathID) {
        room.extendPath(
          pathID,
          selectedColorAndRadiusPoint(...mousePos, pressure),
        )
      } else {
        pathIDsByPointerID.set(
          e.pointerId,
          room.addPath(selectedColorAndRadiusPoint(...mousePos, pressure)),
        )
      }
      if (toolSelection.isRecognitionModeSet()) {
        drawRecognizedUpcoming(room.getPathPoints(pathID), pressure)
      }
      break
    }
    case toolSelection.Tools.ERASER:
      eraseEverythingAtPosition(
        ...mousePos,
        toolSelection.getEraseRadius(),
        room,
      )
      break
    case toolSelection.Tools.DRAGGER:
      dragCanvas(getScreenOffsets(e))
      break
  }
})

window.addEventListener("unload", () => {
  if (room) {
    room.disconnect()
  }
})

// Add the recognition hint element in advance for consistency during tests.
clearRecognizedUpcoming()

tryRoomConnect(initialRoom)
