import { client, xml } from "@xmpp/client"
import uuid from "uuid"

const ChannelState = {
  TRUE: 0,
  FALSE: 1,
  PROCESSING: 4,
}

const SPY_CALLSIGN = "Baguette: " // Vive la France
const GROUP_MESSAGE_ID = "I smell JOJO!" // Ur. Ugly.
const XMPP_STATUS_ROOM_CREATED = "201" // 201 Created

export default class XMPPConnection extends EventTarget {
  joinChannel() {
    const channelIdent = `${this.channel}@${this.details.fqdn}/${this.username}`
    const presence = xml(
      "presence",
      { to: channelIdent },
      xml("x", { xmlns: "http://jabber.org/protocol/muc" }),
    )
    this.sendOrQueue(presence)
  }

  sendOrQueue(message) {
    if (this.online) {
      this.xmpp.send(message)
    } else {
      this.queue.push(message)
    }
  }

  sendChannelOrQueue(message) {
    switch (this.channelState) {
      case ChannelState.TRUE:
        this.sendOrQueue(message)
        break
      case ChannelState.FALSE:
        return
      case ChannelState.PROCESSING:
        this.channelQueue.push(message)
        break
    }
  }

  sendChannelMessage(message) {
    const channelIdent = `${this.channel}@${this.details.fqdn}`
    const wrappedMessage = xml(
      "message",
      {
        type: "groupchat",
        to: channelIdent,
        id: GROUP_MESSAGE_ID,
      },
      xml("body", {}, message),
    )

    this.sendChannelOrQueue(wrappedMessage)
  }

  acceptDefaultRoomConfiguration() {
    const channelIdent = `${this.channel}@${this.details.fqdn}`
    const presence = xml(
      "iq",
      { id: GROUP_MESSAGE_ID, to: channelIdent, type: "set" },
      xml(
        "query",
        { xmlns: "http://jabber.org/protocol/muc#owner" },
        xml("x", { xmlns: "jabber:x:data", type: "submit" }),
      ),
    )
    this.sendOrQueue(presence)
  }

  async processChannelStateChange() {
    const priority = Array.from(this.spyNetwork).sort()
    if (priority[0] === this.username) {
      if (this.channelState === ChannelState.PROCESSING) {
        for (const message of this.channelQueue) {
          await this.sendOrQueue(message)
        }
      }
      this.channelState = ChannelState.TRUE
    } else {
      this.channelState = ChannelState.FALSE
    }
  }

  constructor(channel, details) {
    super()

    this.username = SPY_CALLSIGN + uuid.v4().toString()
    this.channelState = ChannelState.PROCESSING
    this.spyNetwork = new Set()
    this.details = details
    this.channel = channel
    this.channelQueue = []
    this.online = false
    this.queue = []

    this.details.fqdn = "conference." + this.details.host

    const xmpp = client({
      service: `wss://${details.host}:${details.port}/xmpp-websocket`,
      domain: details.host,
      username: details.username,
      password: details.password,
    })

    this.xmpp = xmpp

    xmpp.on("offline", () => {
      this.online = false
    })

    xmpp.on("stanza", (stanza) => {
      const stanzaId = stanza.getAttr("id")
      const delayed = stanza.getChild("delay")
      if (!delayed && stanzaId && stanzaId === GROUP_MESSAGE_ID) {
        // Messages sent to the room as echoed back
        // Ignore our own messages to prevent loops
        // But don't ignore them when we're re-syncing
        return
      }

      if (stanza.is("message")) {
        const body = stanza.getChild("body")
        if (body === undefined) {
          return
        }

        this.dispatchEvent(
          new CustomEvent("stanza", {
            detail: body,
          }),
        )
      } else if (stanza.is("presence")) {
        const from = stanza.getAttr("from")
        if (from === undefined) {
          // Likely won't happen
          return
        }

        const search = `${this.channel}@${this.details.fqdn}/`
        if (from.startsWith(search)) {
          const joiner = from.substring(search.length)
          if (!joiner.startsWith(SPY_CALLSIGN)) {
            return
          }

          const change = stanza.getAttr("type")
          if (change && change === "unavailable") {
            this.spyNetwork.delete(joiner)
          } else {
            this.spyNetwork.add(joiner)
          }

          this.processChannelStateChange()
        }

        const x = stanza.getChild("x")
        if (x === undefined) {
          // Uncertain if this element is guaranteed inside a <presence/>
          return
        }

        const created =
          x.getChildByAttr("code", XMPP_STATUS_ROOM_CREATED) !== undefined
        if (created) {
          // Create an "instant room"
          this.acceptDefaultRoomConfiguration()
        }
      }
    })

    xmpp.on("online", async (address) => {
      /*eslint no-unused-vars: ["error", { "args": "none" }]*/

      // Makes itself available
      await xmpp.send(xml("presence"))

      this.online = true

      for (const message of this.queue) {
        await this.xmpp.send(message)
      }
    })

    xmpp.start().catch(console.error)
    this.joinChannel()
  }

  sneakilySendTheOtherTeamOur(secrets) {
    this.sendChannelMessage(secrets)
  }
}
